/*
 * LEDDriver.c
 *
 * Created: 2015/09/24 12:46:32
 *  Author: NTAKA
 */ 

#include "std.h"

#include <avr/wdt.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>
#include <avr/pgmspace.h>

#include <util/delay.h>

#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include "cmd_parser.h"
#include "rgb.h"

#define EEPROM_VR		0x0000	/* EEPROM uint8_t 0x0000 1Byte */	// R ݒl AhX
#define EEPROM_VG		0x0001	/* EEPROM uint8_t 0x0001 1Byte */	// G ݒl AhX
#define EEPROM_VB		0x0002	/* EEPROM uint8_t 0x0002 1Byte */	// B ݒl AhX
#define EEPROM_INITFLAG	0x0003	/* EEPROM uint8_t 0x0003 1Byte */	//  0 L^wtÕAhX

#define SLOW_INTERVAL	5		// X[X^[g(Xgbv)C^[o[ms]

//#define UART_BAUD		19200	// UART ʐMx
#define UART_BAUD		38400	// UART ʐMx
#define BUFFER_SIZE		32		// ėpIɎgpobt@̃TCY


// I/O
#define IS_PWM_MODE()	((PIND & (1 << PIND7)) ? 1 : 0)	// {[D Ȃ 1
#define IS_PC_POWER()	((PINB & (1 << PINB0)) ? 0 : 1)	// PCd ON Ȃ 0

// PWM l
#define PWM_DEFAULT_R	32
#define PWM_DEFAULT_G	32
#define PWM_DEFAULT_B	32

// ADC ̕σJEg
#define ADC_MEAN_COUNT	16

// PWM ƌ{[ ADC `l
#define PWM_VOLUME_ADC_CH_R	0
#define PWM_VOLUME_ADC_CH_G	1
#define PWM_VOLUME_ADC_CH_B	2

// G[bZ[W
const char ERRMSG1[] PROGMEM = "Buffer over flow!";
const char ERRMSG2[] PROGMEM = "Unkown command!";
const char ERRMSG3[] PROGMEM = "Invalid value!";


// **********************************************************
// ֐`
// **********************************************************

void avr_init_timer0(uint8_t enable);
void avr_init_timer1(uint8_t enable);
void avr_init_timer2(uint8_t enable);

void avr_init();
void avr_soft_reset();

uint8_t get_volume_adc(uint8_t ch);
void set_PWN(const PWMSETTING *setting);

void write_PWM_setting(const PWMSETTING *setting);
void read_PWM_setting(PWMSETTING *setting);


//-----------------------------------------------------------
// EHb`hbO𖳌ɂ
//i 15[ms] ŃEHb`hbN̂łŖĂKv炵j
//-----------------------------------------------------------
uint8_t mcusr_mirror __attribute__ ((section (".noinit")));
void get_mcusr(void) __attribute__((naked)) __attribute__((section(".init3")));

void get_mcusr(void)
{
	mcusr_mirror = MCUSR;
	MCUSR = 0;
	wdt_disable();
}


//-----------------------------------------------------------
// C
//-----------------------------------------------------------
int main(void)
{
	
	
	PWMSETTING comm_pwm_setting = { 0, 0, 0 };		// ʐMŐݒ肳ꂽ PWM ݒl
	PWMSETTING volume_pwm_setting = { 0, 0, 0 };	// {[Őݒ肳ꂽ PWM ݒl
	
	uint8_t pwm_sw_flag = 1;		// PWM  ON/OFF ԃtO
	uint8_t pwm_mode_flag = 0;		// 0 : ʐMD, 1 : {[D
	
	uint8_t pc_power_flag = 0;		// PC p[m̌ʃtO
	uint8_t prev_pc_power_flag = 0;	// OPC p[m̌
	
	uint8_t slow_start_flag = 0;	// X[X^[gtO
	uint8_t slow_stop_flag = 0;		// X[XgbvtO
	
	// |[g
	avr_init();
	
	// EHb`hbNLɂ
	// (t[YrbgWDTON`FbN̂YꂸɁI)
	wdt_enable(WDTO_1S);
	
	// R}hp[T̏
	// (UART ̏s)
	cmd_init(UART_BAUD);
	
	// EEPROM Ǎ
	read_PWM_setting(&comm_pwm_setting);
	
	// 荞݋
	sei();
	
	char recv_buf[BUFFER_SIZE];	// ėpobt@
	char send_buf[BUFFER_SIZE];	// ėpobt@
	
	// ///////////////////////////////////
	// C[v
	while(1)
	{
		// //////////////////////////////////
		// PC dɕωH
		pc_power_flag = IS_PC_POWER();
		
		if(prev_pc_power_flag != pc_power_flag)
		{
			uint8_t tmp[16];
			uint8_t break_flag;
			uint8_t i = 0;
			
			if(pc_power_flag)
			{
				for(uint8_t q = 0 ; q < 16 ; q++)
					tmp[q] = 0;
				
				while(1)
				{
					break_flag = 1;
					
					tmp[i++ & 0x0f] = IS_PC_POWER();
				
					for(uint8_t q = 0 ; q < 16 ; q++)
					{
						if(tmp[q] == 0)
						{
							break_flag = 0;
							break;
						}
					}
					
					if(break_flag)
						break;
					
					_delay_ms(10);
				}	
				
				slow_start_flag = 1;
			}
			else
			{
				for(uint8_t q = 0 ; q < 16 ; q++)
					tmp[q] = 1;
				
				while(1)
				{
					break_flag = 1;
					
					tmp[i++ & 0x0f] = IS_PC_POWER();
					
					for(uint8_t q = 0 ; q < 16 ; q++)
					{
						if(tmp[q] == 1)
						{
							break_flag = 0;
							break;
						}
					}
					
					if(break_flag)
						break;
						
					_delay_ms(10);
				}
				
				slow_stop_flag = 1;
			}
			
			prev_pc_power_flag = pc_power_flag;
		}
		
		
		// //////////////////////////////////
		// {[l̎擾
		volume_pwm_setting.R = get_volume_adc(PWM_VOLUME_ADC_CH_R);
		volume_pwm_setting.G = get_volume_adc(PWM_VOLUME_ADC_CH_G);
		volume_pwm_setting.B = get_volume_adc(PWM_VOLUME_ADC_CH_B);
		
		
		// //////////////////////////////////
		// PWM ύX
		if(slow_start_flag)
		{
			PWMSETTING tmp1 = { 0, 0, 0 };
			const PWMSETTING *tmp2 = (IS_PWM_MODE() || pwm_mode_flag) ? &volume_pwm_setting : &comm_pwm_setting;
			
			avr_init_timer0(1);
			avr_init_timer1(1);
			avr_init_timer2(1);
			
			while(IS_PC_POWER())
			{
				if(tmp1.R < tmp2->R)
					tmp1.R++;
					
				if(tmp1.G < tmp2->G)
					tmp1.G++;
					
				if(tmp1.B < tmp2->B)
					tmp1.B++;
				
				if(tmp1.R >= tmp2->R && tmp1.G >= tmp2->G && tmp1.B >= tmp2->B)
					break;
				
				set_PWN(&tmp1);
				
				_delay_ms(SLOW_INTERVAL);
				
				wdt_reset();
			}
			
			slow_start_flag = 0;
		}
		else if(slow_stop_flag)
		{
			PWMSETTING tmp = (IS_PWM_MODE() || pwm_mode_flag) ? volume_pwm_setting : comm_pwm_setting;
			
			while(!IS_PC_POWER())
			{
				if(tmp.R > 0)
					tmp.R--;
				
				if(tmp.G > 0)
					tmp.G--;
				
				if(tmp.B > 0)
					tmp.B--;
				
				if(tmp.R == 0 && tmp.G == 0 && tmp.B == 0)
					break;
				
				set_PWN(&tmp);
				
				_delay_ms(SLOW_INTERVAL);
				
				wdt_reset();
			}
		
			slow_stop_flag = 0;
			
			avr_init_timer0(0);
			avr_init_timer1(0);
			avr_init_timer2(0);
		}
		else
		{
			if(IS_PWM_MODE() || pwm_mode_flag)
				set_PWN(&volume_pwm_setting);
			else
				set_PWN(&comm_pwm_setting);
		}
		
		// //////////////////////////////////
		// RS232C R}hL[
		
		cmd_paese_from_uart();
		
		while(cmd_is_pop_queue())
		{
			uint8_t res;
			
			enum COMMAND cmd = cmd_pop_queue_all(recv_buf);
			
			switch(cmd)
			{
				case CMD_ERROR:
					cmd_send_str_uart_P(ERRMSG1, SEND_STR_SUFIX_NG);
					break;
					
				case CMD_UNKNOWN:
					cmd_send_str_uart_P(ERRMSG2, SEND_STR_SUFIX_NG);
					break;
				
				
				
				case CMD_GET_R:
					proc_Gx(LEDTYPE_R, &comm_pwm_setting, send_buf);
					cmd_send_str_uart(send_buf, SEND_STR_SUFIX_OK);
					break;
				
				case CMD_GET_G:
					proc_Gx(LEDTYPE_G, &comm_pwm_setting, send_buf);
					cmd_send_str_uart(send_buf, SEND_STR_SUFIX_OK);
					break;
				
				case CMD_GET_B:
					proc_Gx(LEDTYPE_B, &comm_pwm_setting, send_buf);
					cmd_send_str_uart(send_buf, SEND_STR_SUFIX_OK);
					break;
				
				
				
				case CMD_SET_R:
					res = proc_Sx(LEDTYPE_R, recv_buf, &comm_pwm_setting, send_buf);
					cmd_send_str_uart(send_buf, res ? SEND_STR_SUFIX_OK : SEND_STR_SUFIX_NG);
					break;
				
				case CMD_SET_G:
					res = proc_Sx(LEDTYPE_G, recv_buf, &comm_pwm_setting, send_buf);
					cmd_send_str_uart(send_buf, res ? SEND_STR_SUFIX_OK : SEND_STR_SUFIX_NG);
					break;
				
				case CMD_SET_B:
					res = proc_Sx(LEDTYPE_B, recv_buf, &comm_pwm_setting, send_buf);
					cmd_send_str_uart(send_buf, res ? SEND_STR_SUFIX_OK : SEND_STR_SUFIX_NG);
					break;
				
				
				
				case CMD_GET_DATA:
					proc_GD(&comm_pwm_setting, send_buf);
					cmd_send_str_uart(send_buf, SEND_STR_SUFIX_OK);
					break;
					
				case CMD_SET_DATA:
					res = proc_SD(recv_buf, &comm_pwm_setting, send_buf);
					cmd_send_str_uart(send_buf, res ? SEND_STR_SUFIX_OK : SEND_STR_SUFIX_NG);
					break;
					
					
				case CMD_GET_PWM_MODE:
					if(pwm_mode_flag)
						cmd_send_str_uart("GM1", SEND_STR_SUFIX_OK);
					else
						cmd_send_str_uart("GM0", SEND_STR_SUFIX_OK);
					
					break;
					
				case CMD_SET_PWM_MODE:
					if(!str2pwmmode(recv_buf, &pwm_mode_flag))
						cmd_send_str_uart_P(ERRMSG3, 0);
					else
						cmd_send_str_uart(pwm_mode_flag ? "SV1" : "SV0", SEND_STR_SUFIX_OK);
						
					break;
				
				
				
				case CMD_GET_PC_POWER:
					cmd_send_str_uart(pc_power_flag ? "GP1" : "GP0", SEND_STR_SUFIX_OK);	
					break;
				
				
				
				case CMD_SET_PWM_SW:
					if(!str2pwmsw(recv_buf, &pwm_sw_flag))
						cmd_send_str_uart_P(ERRMSG3, SEND_STR_SUFIX_NG);
					else
					{
						avr_init_timer0(pwm_sw_flag);
						avr_init_timer1(pwm_sw_flag);
						avr_init_timer2(pwm_sw_flag);
						
						cmd_send_str_uart("SS", SEND_STR_SUFIX_OK);
					}
					break;
				
				case CMD_GET_PWM_SW:
					cmd_send_str_uart(pwm_sw_flag ? "GS1" : "GS0", SEND_STR_SUFIX_OK);	
					break;
				
				
				
				case CMD_LOAD:
					read_PWM_setting(&comm_pwm_setting);
					cmd_send_str_uart("LD", SEND_STR_SUFIX_OK);
					break;
				
				case CMD_SAVE:
					write_PWM_setting(&comm_pwm_setting);
					cmd_send_str_uart("SV", SEND_STR_SUFIX_OK);
					break;
				
				case CMD_DEFAULT:
					comm_pwm_setting.R = PWM_DEFAULT_R;
					comm_pwm_setting.G = PWM_DEFAULT_G;
					comm_pwm_setting.B = PWM_DEFAULT_B;
					
					cmd_send_str_uart("DT", SEND_STR_SUFIX_OK);
					break;
				
				case CMD_RESET:
					avr_soft_reset();
					break;
				
				default:
					break;
			}
		}
		
		wdt_reset();
	}
}

//-----------------------------------------------------------
// ֐
//-----------------------------------------------------------

void avr_init_timer0(uint8_t enable)
{
	if(enable)
	{
		// PWM OC0As 񔽓]
		TCCR0A = (1 << COM0A1) | (0 << COM0A0);
		TCCR0A |= (1 << WGM01) | (1 << WGM00);
		TCCR0B = (0 << WGM02);
	
		// Ȃ
		TCCR0B |= (0 << CS02) | (0 << CS01) | (1 << CS00);
	}
	else
	{
		TCCR0A = 0;
		TCCR0B = 0;
		
		PORTD &= ~(1 << PORTD6);
	}
	
	OCR0A = 0;
}

void avr_init_timer1(uint8_t enable)
{
	if(enable)
	{
		// 8bit PWM OC1As 񔽓]
		TCCR1A = (1 << COM1A1) | (0 << COM1A0);
		TCCR1A |= (0 << WGM11) | (1 << WGM10);
		TCCR1B = (0 << WGM13) | (1 << WGM12);
		
		// Ȃ
		TCCR1B |= (0 << CS12) | (0 << CS11) | (1 << CS10);
	}
	else
	{
		TCCR1A = 0;
		TCCR1B = 0;
		
		PORTB &= ~(1 << PORTB1);
	}
	
	OCR1AL = 0;
}

void avr_init_timer2(uint8_t enable)
{
	if(enable)
	{
		// PWM OC2As 񔽓]
		TCCR2A = (1 << COM2A1) | (0 << COM2A0);
		TCCR2A |= (1 << WGM21) | (1 << WGM20);
		TCCR2B = (0 << WGM22);
		
		// Ȃ
		TCCR2B |= (0 << CS22) | (0 << CS21) | (1 << CS20);
	}
	else
	{
		TCCR2A = 0;
		TCCR2B = 0;
		
		PORTB &= ~(1 << PORTB3);
	}
	
	OCR2A = 0;
}

void avr_init()
{
	//////////////////////////////////////
	// s̓o͐ݒ
	
	//		PB	76543210
	PORTB	= 0b00000001;	// ͂̏ꍇ 1 : vAbvA 0 : vAbvȂB o͂̏ꍇ  0, 1 s̏o
	DDRB	= 0b00001010;	// 0 : ́A1 : o
	
	// PB0 in  PCdON m
	// PB1 out ^C}[1 PWM
	// PB3 out ^C}[2 PWM
	
	//		PD	76543210
	PORTD	= 0b10000000;	// ͂̏ꍇ 1 : vAbvA 0 : vAbvȂB o͂̏ꍇ  0, 1 s̏o
	DDRD	= 0b01000010;	// 0 : ́A1 : o
	
	// PD0 in  USART Rx
	// PD1 out USART Tx
	
	// PD6 out ^C}[0 PWM
	// PD7 in  }jA PWM ON
	
	//////////////////////////////////////
	// A/D ̐ݒ
	
	// fW^͋֎~
	DIDR0 = (1 << ADC2D) | (1 << ADC1D) | (1 << ADC0D);
	
	// ADC , 128(62.5kHz)
	ADCSRA = (1 << ADEN) | (1 << ADPS2) | (1 << ADPS1) | (1 << ADPS0);
}

//-----------------------------------------------------------
// AVR ̃\tgZbg
//-----------------------------------------------------------
void avr_soft_reset()
{
	wdt_enable(WDTO_15MS);
	
	while(1)
		;
}

//-----------------------------------------------------------
// ADC ̎s
//-----------------------------------------------------------
uint8_t get_volume_adc(uint8_t ch)
{
	uint16_t sum = 0;
	
	// d AVcc, , }`vNT̐ݒ, ADC ̃`lݒ
	ADMUX = (0 << REFS1) | (1 << REFS0) | (1 << ADLAR) | ch;
	
	for(uint8_t i = 0 ; i < ADC_MEAN_COUNT ; i++)
	{
		// ϊJn
		ADCSRA &= ~(1 << ADIF);
		ADCSRA |= (1 << ADSC);
		
		// ϊ҂
		loop_until_bit_is_set(ADCSRA, ADIF);
		
		sum += (uint16_t)ADCH;
	}
	
	return (uint8_t)(sum / ADC_MEAN_COUNT);
}

//-----------------------------------------------------------
// PWM l̐ݒ
//-----------------------------------------------------------
void set_PWN(const PWMSETTING *setting)
{
	OCR0A  = setting->R;
	OCR1AL = setting->G;
	OCR2A  = setting->B;
}

//-----------------------------------------------------------
// PWM l̏
//-----------------------------------------------------------
void write_PWM_setting(const PWMSETTING *setting)
{
	eeprom_busy_wait();
	eeprom_write_byte((uint8_t *)EEPROM_VR, setting->R);
	wdt_reset();
	
	eeprom_busy_wait();
	eeprom_write_byte((uint8_t *)EEPROM_VG, setting->G);
	wdt_reset();
	
	eeprom_busy_wait();
	eeprom_write_byte((uint8_t *)EEPROM_VB, setting->B);
	wdt_reset();
	
	eeprom_busy_wait();
	eeprom_write_byte((uint8_t *)EEPROM_INITFLAG, 0);
	wdt_reset();
}

//-----------------------------------------------------------
// PWM l̓ǂݍ
//-----------------------------------------------------------
void read_PWM_setting(PWMSETTING *setting)
{
	eeprom_busy_wait();
	setting->R = eeprom_read_byte((uint8_t *)EEPROM_VR);
	wdt_reset();
	
	eeprom_busy_wait();
	setting->G = eeprom_read_byte((uint8_t *)EEPROM_VG);
	wdt_reset();
	
	eeprom_busy_wait();
	setting->B = eeprom_read_byte((uint8_t *)EEPROM_VB);
	wdt_reset();
	
	eeprom_busy_wait();
	uint8_t flag = eeprom_read_byte((uint8_t *)EEPROM_INITFLAG);
	wdt_reset();
	
	
	// H
	if(flag != 0)
	{
		setting->R = PWM_DEFAULT_R;
		setting->G = PWM_DEFAULT_G;
		setting->B = PWM_DEFAULT_B;
	}
}