/*
 * rgb.c
 *
 * Created: 2015/09/29 15:43:03
 *  Author: NTAKA
 */ 

#include "std.h"

#include <string.h>
#include <stdlib.h>

#include <avr/pgmspace.h>

#include "rgb.h"

#include "helper.h"


const char error_msg1[] PROGMEM = "invalid value param1!";
const char error_msg2[] PROGMEM = "invalid value param2!";
const char error_msg3[] PROGMEM = "invalid value param3!";
const char error_msg9[] PROGMEM = "internal error!";


uint8_t proc_Gx(uint8_t led_type, const PWMSETTING *pwm,  char *dst)
{
	switch(led_type)
	{
	case LEDTYPE_R:
		strcpy(dst, "GRvv");
		replace_val2str_2(pwm->R, dst, 2);
		break;
		
	case LEDTYPE_G:
		strcpy(dst, "GGvv");
		replace_val2str_2(pwm->G, dst, 2);
		break;
		
		
	case LEDTYPE_B:
		strcpy(dst, "GBvv");
		replace_val2str_2(pwm->B, dst, 2);
		break;
		
	default:
		strcpy_P(dst, error_msg9);
		return 0;
	}
	
	return 1;
}

uint8_t proc_Sx(uint8_t led_type, const char *str, PWMSETTING *pwm, char *dst)
{
	uint8_t val;
	
	if(!str2val_2(&val, str, 2))
	{
		strcpy_P(dst, error_msg1);
		return 0;
	}
	
	switch(led_type)
	{
	case LEDTYPE_R:
		pwm->R = val;
		strcpy(dst, "SR");
		break;
	
	case LEDTYPE_G:
		pwm->G = val;
		strcpy(dst, "SG");
		break;
		
	
	case LEDTYPE_B:
		pwm->B = val;
		strcpy(dst, "SG");
		break;
		
	default:
		strcpy_P(dst, error_msg9);
		return 0;
	}
	
	return 1;
}

void proc_GD(const PWMSETTING *pwm, char *dst)
{
	strcpy(dst, "GDrr,gg,bb");
	
	replace_val2str_2(pwm->R, dst, 2);
	replace_val2str_2(pwm->G, dst, 5);
	replace_val2str_2(pwm->B, dst, 8);
}

// SD : set data R}h
uint8_t proc_SD(const char *str, PWMSETTING *pwm, char *dst)
{
	// 0123456789
	// SDrr,gg,bb
	
	PWMSETTING tmp = { 0 };
	
	if(!str2val_2(&tmp.R, str, 2))
	{
		strcpy_P(dst, error_msg1);
		return 0;
	}
	
	if(!str2val_2(&tmp.G, str, 5))
	{
		strcpy_P(dst, error_msg2);
		return 0;
	}
	
	if(!str2val_2(&tmp.B, str, 8))
	{
		strcpy_P(dst, error_msg3);
		return 0;
	}
	
	*pwm = tmp;
	
	strcpy(dst, "SD");
	
	return 1;
}

uint8_t str2pwmmode(const char *str, uint8_t *mode)
{
	uint8_t val;
	
	if(!str2val_1(&val, str, 2))
		return 0;
		
	*mode = val;
	
	return 1;
}

uint8_t str2pwmsw(const char *str, uint8_t *sw)
{
	uint8_t val;
	
	if(!str2val_1(&val, str, 2))
	return 0;
	
	*sw = val;
	
	return 1;
}


