/*
 * usart.c
 *
 * Created: 2015/04/09 16:22:14
 *  Author: NTAKA
 */ 

#include "std.h"

#include <avr/interrupt.h>

#include "usart.h"

#define TXIE_ENABLE_()		UCSR0B |= (1 << UDRIE0)
#define TXIE_DISABLE_()		UCSR0B &= ~(1 << UDRIE0)
#define RXCIE_ENABLE_()		UCSR0B |= (1 << RXCIE0)
#define RXCIE_DISABLE_()	UCSR0B &= ~(1 << RXCIE0)


#define BUFFER_SIZE_	64

typedef struct tag_DATABUF
{
	char buf[BUFFER_SIZE_];	// obt@
	int wpos;				// ݂writeʒu(݃CfbNX)
	int rpos;				// ݂readʒu(Ǎ݃CfbNX)
} DATABUF;

volatile DATABUF g_recv;
volatile DATABUF g_send;


//-----------------------------------------------------------
// USART 
//-----------------------------------------------------------
void uart_init(uint16_t baud)
{
	g_recv.rpos = g_recv.wpos = 0;
	g_send.rpos = g_send.wpos = 0;
	
	uint16_t ubrr = (((F_CPU>>4) + (baud>>1)) / baud - 1);
	UBRR0H = (unsigned char)(ubrr>>8);	// {[[g8bit
	UBRR0L = (unsigned char)ubrr;		// {[[g8bit
	
	// M,M荞݋
	UCSR0B = (1 << RXEN0) | (1 << TXEN0) | (1 << RXCIE0);
	
	// 񓯊 ()Af[^8rbgApeBȂAXgbv1rbg
	UCSR0C = (1 << UCSZ01) | (1 << UCSZ00);
}


//-----------------------------------------------------------
// f[^MĂ邩̃`FbNBMĂȂ0B
//-----------------------------------------------------------
int uart_is_received()
{
	return  (g_recv.wpos != g_recv.rpos) ? 1 : 0;
}

//-----------------------------------------------------------
// f[^M܂őҋ@
//-----------------------------------------------------------
void uart_wait_for_receiving()
{
	while(!uart_is_received())
		;
}

//-----------------------------------------------------------
// Mf[^ԂBMf[^Ȃꍇ͎M܂őҋ@B
//-----------------------------------------------------------
char uart_recv_byte()
{
	uart_wait_for_receiving();
	
	// M荞݂񋖉
//	_RXCIE_DISABLE();
	
	char c = g_recv.buf[g_recv.rpos++];
	
	if(g_recv.rpos >= BUFFER_SIZE_)
		g_recv.rpos = 0;
	
	// M荞݋
//	_RXCIE_ENABLE();
	
	return c;
}

//-----------------------------------------------------------
// 荞݂ɂM
//-----------------------------------------------------------
ISR(USART_RX_vect)
{
	g_recv.buf[g_recv.wpos++] = UDR0;
	
	if(g_recv.wpos >= BUFFER_SIZE_)
		g_recv.wpos = 0;
}

//-----------------------------------------------------------
// 荞݂ɂ鑗M
//-----------------------------------------------------------
ISR(USART_UDRE_vect)
{
	// Sf[^̑MIꍇ́AuMWX^󂫊荞݁v̋֎~đMI
	if(g_send.wpos == g_send.rpos)
	{
		TXIE_DISABLE_();
	}
	else
	{
		UDR0 = g_send.buf[g_send.rpos++];
	
		if(g_send.rpos >= BUFFER_SIZE_)
			g_send.rpos = 0;
	}
}

//-----------------------------------------------------------
// 1oCgM\𒲂ׂ
//-----------------------------------------------------------
int uart_is_send()
{
	if(g_send.wpos == (BUFFER_SIZE_ - 1))
	{
		return (g_send.rpos == 0) ? 0 : 1;
	}
	
	return ((g_send.wpos + 1) == g_send.rpos) ? 0 : 1;
}

//-----------------------------------------------------------
// 1oCgM
//-----------------------------------------------------------
void uart_send_byte(char c)
{
	// Mobt@ςȂ҂
	while(!uart_is_send())
		;
	
	// uMWX^󂫊荞݁v̋֎~
	TXIE_DISABLE_();
	
	// Mobt@Ƀf[^ςށB
	g_send.buf[g_send.wpos++] = c;
	
	if(g_send.wpos >= BUFFER_SIZE_)
		g_send.wpos = 0;
	
	// M̃f[^Ȃ΁AuMWX^󂫊荞݁vđM
	if(g_send.wpos != g_send.rpos)
		TXIE_ENABLE_();
}

//-----------------------------------------------------------
// ̑M
//-----------------------------------------------------------
void uart_send_string(const char* str)
{
	while(*str)
		uart_send_byte(*str++);
}
